==> plantron-1.4.4.tar.gz: plantron/agentes-aluXXXX-camino.lisp <==
;; Programa: Package con agentes jugadores buscadores de
;;           camino PlanTron mediante sistemas de producción
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: 3 (Sistemas de Producción II)
;; Autor:    aluXXXX ([Indica aquí tu nombre completo])
;; Fecha:    26/04/2010
;; Comentarios:
;;
;;           Este código supone un esqueleto de ejemplo de el
;;           fichero que ha de contener los agentes jugadores
;;           de PlanTron que el usuario defina y usen búsqueda
;;           en sistemas de producción.
;;
;;           Para personalizar este fichero ha de renombrarse,
;;           sustituyendo "aluXXXX" por el lo que desee. Además
;;           en este código ha de reemplazarse dondequiera que
;;           apareza "aluXXXX" o "ALUXXXX" por lo mismo, ya que
;;           el nombre del fichero sin extensión y el nombre
;;           del package aquí incluido han de coincidir (no
;;           sensitivo a mayúsculas).
;;
;;           Para añadir más agentes habrán de generarse las
;;           funciones correspondientes para los agentes, e 
;;           incluirlas como nuevo agente mediante el uso de
;;           la macro DEFAGENTE. Para más información consultar
;;           ejemplos de "pt-agentes-simples.lisp" y la documentación
;;           (DESCRIBE) de las constantes, variables y funciones allí
;;           utilizadas.
;;
;;           Los agentes definidos utilizan el entorno de juego
;;           PlanTron, definido principalmente por los packages
;;           PT, PT-BASE y ENVIO.
;;
;;           Las funciones accesibles más importantes son:
;;
;;             ALUXXXX-CAMINO-BA: Agente PlanTron que busca
;;             el camino a la posición del adversario mediante
;;             búsqueda en anchura
;;
;;             ALUXXXX-CAMINO-A: Agente PlanTron que busca
;;             el camino a la posición del adversario mediante
;;             búsqueda en A*
;;
;;           Ver la documentación existente para cada una de ellas.
;;
;;           Para su utilización es aconsejable usar el cargador
;;           del entorno de juego PlanTron: "pt-loader.lisp".
;;
;;
;; Dependencias:
;;
;;           Hace uso del paquete que contiene la interface (ENVIO)
;;           con el control del juego (PT-BASE).
;;
;;           Hace uso del paquete PT-UTIL con agentes pasivos
;;           con distintas utilidades para asistir a agentes jugadores
;;           de PlanTron (agente pasivo CONTADOR) e interfaces para
;;           acceso a paquetes SHELL, GENERIC-SEACH y A_STAR.
;;
;;           Hace uso del paquete interfaz para la creación de sistemas
;;           de producción simples (SHELL) a través de PT-UTIL.
;;
;;           Hace uso de la función AGENTE-ESPIRAL y AGENTE-COBARDE del
;;           paquete PT-AGENTES-SIMPLES.
;;
;;           Hace uso del paquete marco genérico para la búsquedas en
;;           sistemas de producción simples (GENERIC-SEARCH) a través
;;           de PT-UTIL.
;;
;;           Hace uso del paquete para la búsquedas A* (A_STAR) en
;;           sistemas de producción simples a través de PT-UTIL
;;
;;           Hace uso inderecto del paquete PT-VIS mediante la
;;           solicitud de impresión del camino encontrado en la
;;           función PRINT-CAMINO.
;;
;;
;; Modificaciones:
;;
;;           - Ver 0.1 (27/04/2010): Primera versión
;;
;;
;; Por hacer:
;;
;;
;;           - Mejoras en funcionalidades
;;             --------------------------
;;           - Cuando la búsqueda emplea el tiempo disponible sin 
;;             haber alcanzado un objetivo usar camino a lugar  
;;             encontrado más cercano al adversario (Dificultad:  
;;             Fácil; Mejora esperada: Poca)
;;           - Detectar cuando el agente está ya muy próximo al  
;;             adversario y actuar de otra forma más adecuada para  
;;             intentar eliminarlo (Dificultad: Fácil / Media;  
;;             Mejora esperada: Poca)
;;           - Cuando se detecte agente encerrado, cambiar  
;;             estrategia a buscar camino muy largo de casillas  
;;             vacías (Dificultad: Media; Mejora esperada: Media)
;;           - Codificar otra estrategia de búsqueda heurística  
;;             distinta al A* (Dificultad: Media; Mejora  
;;             esperada: Media)
;;           - Implementar tratamiento de estados visitados  
;;             (Dificultad: Media / Difícil; Mejora esperada:  
;;             Media / Alta)
;;           - Cambiar radicalmente de estrategia, buscando caminos  
;;             que lo guíen a zonas alejadas a una determinada  
;;             distancia del adversario (Dificultad: Media /  
;;             Difícil; Mejora esperada: Media / Alta)
;;           - Reutilizar la base de hechos de búsquedas en  
;;             anteriores turnos en nuevas búsquedas (Dificultad:  
;;             Difícil / Muy difícil; Mejora esperada: Media)
;;
;;


;;
;; Definición del package
;; ----------------------

(defpackage "AGENTES-ALUXXXX-CAMINO"
  (:use "COMMON-LISP")
  (:documentation "Agentes jugadores de PlanTron basado en búsquedas.")
  (:export "VERSION"))

(in-package "AGENTES-ALUXXXX-CAMINO")


;;
;; Constante versión
;; -----------------

(defconstant version "0.1"
  "Versión del package AGENTES-ALUXXXX-CAMINO.")


;;
;; Definición de estructura de estados
;; -----------------------------------

(defstruct estado
  "Almacena los estados del sistema de producción para presas."
  (camino nil)       ; Camino como lista inversa de posiciones (x y)
  (lon-camino 0)     ; Longitud del camino
  (f-heur 0)         ; Valor de la función heurística f'
  (dir-ini nil))     ; Dirección inicial del camino: [nil, 0, 1, 2 ó 3]


;;
;; Definición de varialbes globales
;; --------------------------------

(defvar *f-heur* nil
  "Función heurística que evalua un estado en acción de regla.")

(defvar pos-adv nil
  "Posición donde está el adversario")


;;
;; Funciones de apoyo para elaboración de reglas
;; ---------------------------------------------

(defun pos-posible-p (pos &optional dir)
  "Mira si es posible colocarse en la posición resultante de moverse en la dirección DIR a la posición POS. Devuelve algo verdadero en tal caso y NIL en caso contrario. POS es una posición del tablero dada por una lista con valor de coordenada X y valor de coordenada Y. DIR ha de ser un número entero (dirección) entre 0 y 3 (norte, este, sur y oeste respectivamente) o NIL si se quiere consultar la posición POS."
  (eq (envio:consulta-tipo (envio:consulta-pos-tab pos dir)) nil))

(defconstant dir-inc
  (make-array '(4 2)
	      :initial-contents
	      '((1 -1)    ; Norte
		(0 +1)    ; Este
		(1 +1)    ; Sur
		(0 -1)))  ; Oeste
  "Array para transformar de dirección a número de coordenada a modificar y valor de incremento.")

(defmacro dir2dim-inc (dir)
  `(aref dir-inc ,dir 0))

(defmacro dir2val-inc (dir)
  `(aref dir-inc ,dir 1))

(defun pos-inc-dir (pos dir)
  "Calcula nueva lista con la posición resultante de desplazarse en la dirección DIR respecto a POS."
  (let ((pos-inc (copy-list pos)))
    (setf (nth (dir2dim-inc dir) pos-inc)
	  (+ (nth (dir2dim-inc dir) pos-inc)
	     (dir2val-inc dir)))
    pos-inc))

(defun distancia-city (pos1 pos2)
  "Calcula la distancia city block entre POS1 y POS2."
  (+ (abs (- (first pos1) (first pos2)))
     (abs (- (second pos1) (second pos2)))))

(defun dist-adv (pos)
  "Calcula la distancia city block desde POS hasta donde está el adversario. Si no existe adversario devuelve NIL."
  (if pos-adv
      (distancia-city pos pos-adv)))


;;
;; Funciones para generar sistema de producción
;; --------------------------------------------

(defun estados-iniciales (id)
  "Genera lista de estados iniciales del jugador con identificador ID."
  (let ((nue-est (make-estado :camino (list (envio:consulta-pos-jug id)))))
    (if *f-heur*
	(setf (estado-f-heur nue-est) (funcall *f-heur* nue-est)))
    (list nue-est)))

(defun precond-regla-dir (est dir)
  "Generador de precondicion de regla de mover el estado EST en dirección DIR."
  (let ((pos-est (first (estado-camino est))))
    ; La posición no pertenece al propio camino
    (and (not (member (pos-inc-dir pos-est dir)
		      (rest (estado-camino est))
		      :test #'equal))
	 ; Y la posición se puede pisar
	 (pos-posible-p pos-est dir))))

(defun accion-regla-dir (est dir)
  "Generador de acción de regla de mover el estado EST en dirección DIR."
  (let ((nue-est (make-estado :camino (cons (pos-inc-dir (first (estado-camino est)) dir)
					    (estado-camino est))
			      :lon-camino (1+ (estado-lon-camino est))
			      :dir-ini (or (estado-dir-ini est) dir))))
    ; Asignar valor heurísticos, si se requiere
    (if *f-heur*
	(setf (estado-f-heur nue-est) (funcall *f-heur* nue-est)))
    nue-est))

(defun print-camino (camino)
  "Solicita imprimir la lista de posiciones dada en CAMNIO a los agentes capaces de acpetar solicitudes tipo :punteo."
  (envio:broadcast :punteo (reverse camino)))


;;
;; Inicializador de agentes de búsqueda de camino
;; ----------------------------------------------

(defun aluXXXX-camino-ini (id tipo-juego)
  "Inicialización genérica de agente que buscan camino."
  (declare (ignore id tipo-juego))          ; Ignorar parámetros no usados
  (if (pt-util:usa-sistema 'camino-presa)   ; Si no existían ya las reglas
      (progn
	; Insertar las nuevas reglas
	(pt-util:regla #'(lambda (est) (precond-regla-dir est 3))  ; Regla Oeste
		       #'(lambda (est) (accion-regla-dir est 3)))
	(pt-util:regla #'(lambda (est) (precond-regla-dir est 1))  ; Regla Este
		       #'(lambda (est) (accion-regla-dir est 1)))
	(pt-util:regla #'(lambda (est) (precond-regla-dir est 0))  ; Regla Norte
		       #'(lambda (est) (accion-regla-dir est 0)))
	(pt-util:regla #'(lambda (est) (precond-regla-dir est 2))  ; Regla Sur
		       #'(lambda (est) (accion-regla-dir est 2)))))
  t)    ; Devolver T siempre


;;
;; Agente búsqueda de camino en anchura
;; ------------------------------------

(defun estado-objetivo-p (est)
  "Chequea si un estado EST es estado objetivo o NIL."
  ; Si se ha acabado el tiempo
  (or (pt-util:consulta-contador)
      ; o no se ha encontrado solución
      (null est)
      ; o no hay adversario
      (null pos-adv)
      ; o ya se ha contactado con el adversario
      (= (dist-adv (first (estado-camino est))) 1)))

(defun aluXXXX-camino-ba (id)
  "Agente jugador PlanTron que busca camino en anchura a lugar lejos de depredador."
  ; Iniciar contador de tiempo transcurrido
  (pt-util:inicia-contador)
  ; Usar sistema de producción de la presa
  (pt-util:usa-sistema 'camino-presa)
  ; Cargar heurística a utilizar (sin heurística)
  (setf *f-heur* nil)
  ; Establecer posición del adversario (devuelve nil si no existe)
  (setf pos-adv (envio:consulta-pos-jug (if (eq id 1) 2 1)))
  ; Lanzar búsqueda
  (let ((est-sol
	 (pt-util:generic-search
	  (estados-iniciales id)
	  #'estado-objetivo-p
	  #'pt-util:hijos
	  :display-fn #'null
	  :merge-fn   #'(lambda (new-states)
			  (nconc pt-util:*search-queue* new-states))
	  :dequeue-fn #'(lambda () (pop pt-util:*search-queue*)))))
    ; Preguntar por condiciones de parada
    (cond
     ((null est-sol)                   ; No encontrada solución (encerrado)
      (envio:mensaje "Encerrado :( lanzo espiral")
      (pt-agentes-simples::agente-espiral id))   ; Como agente espiral

     ((null pos-adv)                   ; No existe adversario
      (envio:mensaje "No detectado adversario :) lanzo espiral")
      (pt-agentes-simples::agente-espiral id))   ; Como agente espiral
     
     ((pt-util:consulta-contador)      ; Agotado el tiempo de búsqueda
      (envio:mensaje "Agotado el tiempo de búsqueda")
      (estado-dir-ini est-sol))                  ; Mover según estado actual

     ((null (estado-dir-ini est-sol))  ; Ya pegado al adversario
      (envio:mensaje "Pegado al adversario :D lanzo cobarde")
      (pt-agentes-simples::agente-cobarde id))   ; Como agente cobarde

     (t                                ; Encontrado camino
      (envio:mensaje "Encontrado camino BA")
      (print-camino (estado-camino est-sol))
      (estado-dir-ini est-sol)))))               ; Mover según estado solución


;;
;; Agente búsqueda de camino A*
;; ----------------------------

(defun f-dist-adv (est)
  "Función heurística subdividida con coste del camino acumulado + distancia estimada por city block que falta para llegar a contactar al adversario."
  (let ((dt (dist-adv (first (estado-camino est)))))
    (+ (estado-lon-camino est)
       (if dt (- dt 1) 0))))   ; Si existe adversario suma distancia hasta él
	                       ; sino suma 0

(defun estado-objetivo-a-p (est)
  "Chequea si un estado EST es estado objetivo o NIL. Aprovecha el campo F-HEUR del estado."
  ; Si se ha acabado el tiempo
  (or (pt-util:consulta-contador)
      ; o no se ha encontrado solución
      (null est)
      ; o ya encontramos el camino
      (= (estado-lon-camino est) (estado-f-heur est))))

(defun igual-estado (est1 est2)
  "Comprueba si el estado EST1 es igual al EST2"
  ; Considerar dos estados iguales si su camino
  ; acaba en la misma posición
  (equal (first (estado-camino est1))
         (first (estado-camino est2))))

(defun aluXXXX-camino-a (id)
  "Agente jugador PlanTron que busca con algoritmo A camino a lugar lejos de depredador."
  ; Iniciar contador de tiempo
  (pt-util:inicia-contador)
  ; Usar sistema de producción de la presa
  (pt-util:usa-sistema 'camino-presa)
  ; Inicializar variables para heurísticas del A*
  (setf *f-heur* #'f-dist-adv)                        ; Fun. f' para reglas
  (setf pt-util:*coste-acum-fn* #'estado-lon-camino)  ; Fun. g  para búsqueda
  (setf pt-util:*eval-fn* #'estado-f-heur)            ; Fun. f' para búsqueda
  (setf pt-util:*igual-est-fn* #'igual-estado)        ; Pred. estados iguales
  ; Establecer posición del adversario (devuelve nil si no existe)
  (setf pos-adv (envio:consulta-pos-jug (if (eq id 1) 2 1)))
  ; Lanzar búsqueda
  (let ((est-sol
	 (pt-util:generic-search
	  (estados-iniciales id)
	  #'estado-objetivo-a-p
	  #'pt-util:hijos
	  :display-fn #'null
	  :merge-fn   #'pt-util:priority-a-merge-fn
	  :dequeue-fn #'pt-util:priority-a-dequeue-fn)))
    ; Preguntar por condiciones de parada
    (cond
     ((null est-sol)                   ; No encontrada solución (encerrado)
      (envio:mensaje "Encerrado :( lanzo espiral")
      (pt-agentes-simples::agente-espiral id))   ; Como agente espiral

     ((null pos-adv)                   ; No existe adversario
      (envio:mensaje "No detectado adversario :) lanzo espiral")
      (pt-agentes-simples::agente-espiral id))   ; Como agente espiral
     
     ((pt-util:consulta-contador)      ; Agotado el tiempo de búsqueda
      (envio:mensaje "Agotado el tiempo de búsqueda")
      (estado-dir-ini est-sol))                  ; Mover según estado actual

     ((null (estado-dir-ini est-sol))  ; Ya pegado al adversario
      (envio:mensaje "Pegado al adversario :D lanzo cobarde")
      (pt-agentes-simples::agente-cobarde id))   ; Como agente cobarde

     (t                                ; Encontrado camino
      (envio:mensaje "Encontrado camino A*")
      (print-camino (estado-camino est-sol))
      (estado-dir-ini est-sol)))))               ; Mover según estado solución


;;
;; Definiciones de agentes presentes en este package
;; -------------------------------------------------

(pt:defagente aluXXXX-camino-ba
	      :turno #'aluXXXX-camino-ba
	      :inicio #'aluXXXX-camino-ini
	      :use '("CONTADOR" "SHELL" "GENERIC-SEARCH")
	      :documentation "Agente jugador PlanTron que busca camino en anchura al adversario.")

(pt:defagente aluXXXX-camino-a
	      :turno #'aluXXXX-camino-a
	      :inicio #'aluXXXX-camino-ini
	      :use '("CONTADOR" "SHELL" "GENERIC-SEARCH" "A_STAR")
	      :documentation "Agente jugador PlanTron que busca camino con algoritmo A al adversario.")

==> plantron-1.5.2.tar.gz: plantron/agentes-aluXXXX-camino.lisp <==
;; Programa: Package con agentes jugadores presa de PlanTron
;;           contra depredadores basados en búsquedas con
;;           sistemas de producción
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: 3 (Sistemas de Producción I)
;; Autor:    aluXXXX ([Indica aquí tu nombre completo])
;; Fecha:    26/04/2012
;; Comentarios:
;;
;;           Este código supone un esqueleto de ejemplo de el
;;           fichero que ha de contener los agentes jugadores
;;           de PlanTron que el usuario defina y usen búsqueda
;;           en sistemas de producción.
;;
;;           Para personalizar este fichero ha de renombrarse,
;;           sustituyendo "aluXXXX" por el lo que desee. Además
;;           en este código ha de reemplazarse dondequiera que
;;           apareza "aluXXXX" o "ALUXXXX" por lo mismo, ya que
;;           el nombre del fichero sin extensión y el nombre
;;           del package aquí incluido han de coincidir (no
;;           sensitivo a mayúsculas).
;;
;;           Para añadir más agentes habrán de generarse las
;;           funciones correspondientes para los agentes, e 
;;           incluirlas como nuevo agente mediante el uso de
;;           la macro DEFAGENTE. Para más información consultar
;;           ejemplos de "pt-agentes-simples.lisp" y la documentación
;;           (DESCRIBE) de las constantes, variables y funciones allí
;;           utilizadas.
;;
;;           Los agentes definidos utilizan el entorno de juego
;;           PlanTron, definido principalmente por los packages
;;           PT, PT-BASE y ENVIO.
;;
;;           Las funciones accesibles más importantes son:
;;
;;             ALUXXXX-CAMINO-BA: Agente PlanTron que busca
;;             el camino a planta mediante búsqueda en anchura
;;
;;             ALUXXXX-CAMINO-A: Agente PlanTron que busca
;;             el camino a planta mediante búsqueda en A*
;;
;;           Ver la documentación existente para cada una de ellas.
;;
;;           Para su utilización es aconsejable usar el cargador
;;           del entorno de juego PlanTron: "pt-loader.lisp".
;;
;;
;; Dependencias:
;;
;;           Hace uso del paquete que contiene la interface (ENVIO)
;;           con el control del juego (PT-BASE).
;;
;;           Hace uso del paquete PT-UTIL con agentes pasivos
;;           con distintas utilidades para asistir a agentes jugadores
;;           de PlanTron (agente pasivo CONTADOR) e interfaces para
;;           acceso a paquetes SHELL, GENERIC-SEACH y A_STAR.
;;
;;           Hace uso del paquete interfaz para la creación de sistemas
;;           de producción simples (SHELL) a través de PT-UTIL.
;;
;;           Hace uso de la función AGENTE-ESPIRAL del paquete
;;           PT-AGENTES-SIMPLES.
;;
;;           Hace uso del paquete marco genérico para la búsquedas en
;;           sistemas de producción simples (GENERIC-SEARCH) a través
;;           de PT-UTIL.
;;
;;           Hace uso del paquete para la búsquedas A* (A_STAR) en
;;           sistemas de producción simples a través de PT-UTIL
;;
;;           Hace uso inderecto del paquete PT-VIS mediante la
;;           solicitud de impresión del camino encontrado en la
;;           función PRINT-CAMINO.
;;
;;
;; Modificaciones:
;;
;;           - Ver 0.1 (26/04/2012): Primera versión
;;
;;
;; Por hacer:
;;
;;
;;           - Posibles mejoras en funcionalidades a implementar
;;             -------------------------------------------------
;;           - Cuando la búsqueda emplea el tiempo disponible sin
;;             haber alcanzado un objetivo usar camino a lugar
;;             encontrado más cercano a una planta (Dificultad:
;;             Fácil; Mejora esperada: Media)
;;           - Cuando agentes está cerca del depredador cambiar de
;;             estrategia y mover con otro agente (Dificultad:
;;             Fácil; Mejora esperada: Poca / Media)
;;           - Preferir caminos que pasen menos cerca del depredador
;;             (Dificultad: Media; Mejora esperada: Poca / Media)
;;           - Codificar otra estrategia de búsqueda heurística
;;             distinta al A* (Dificultad: Media; Mejora esperada:
;;              Media)
;;           - Implementar tratamiento de estados visitados
;;             (Dificultad: Media / Difícil; Mejora esperada: Media)
;;           - Utilizar información de varias plantas cercanas para
;;             guiar la búsqueda con A* (Dificultad: Media; Mejora
;;             esperada: Media)
;;           - Buscar con A* caminos de longitud fija que pasen
;;             sobre más plantas (Dificultad: Media / Difícil;
;;             Mejora esperada: Media / Alta)
;;           - Cuando agente está cerca del depredador moverse según
;;             resultado de búsqueda a lugar alejado del depredador
;;             (Dificultad: Media / Difícil; Mejora esperada:
;;             Media / Alta)
;;           - Reutilizar la base de hechos de búsquedas en
;;             anteriores turnos en nuevas búsquedas (Dificultad:
;;             Difícil / Muy difícil; Mejora esperada: Media / Alta)
;;
;;


;;
;; Definición del package
;; ----------------------

(defpackage "AGENTES-ALUXXXX-CAMINO"
  (:use "COMMON-LISP")
  (:documentation "Agentes jugadores de PlanTron basado en búsquedas.")
  (:export "VERSION"))

(in-package "AGENTES-ALUXXXX-CAMINO")


;;
;; Constante versión
;; -----------------

(defconstant version "0.1"
  "Versión del package AGENTES-ALUXXXX-CAMINO.")


;;
;; Definición de estructura de estados
;; -----------------------------------

(defstruct estado
  "Almacena los estados del sistema de producción para presas."
  (camino nil)       ; Camino como lista inversa de posiciones (x y)
  (lon-camino 0)     ; Longitud del camino
  (f-heur 0)         ; Valor de la función heurística f'
  (dir-ini nil))     ; Dirección inicial del camino: [nil, 0, 1, 2 ó 3]


;;
;; Definición de varialbes globales
;; --------------------------------

(defvar *f-heur* nil
  "Función heurística que evalua un estado en acción de regla.")


;;
;; Funciones de apoyo para elaboración de reglas
;; ---------------------------------------------

(defun pos-posible-p (pos &optional dir)
  "Mira si es posible colocarse (planta o vacío) en la posición resultante de moverse en la dirección DIR a la posición POS. Devuelve algo verdadero en tal caso y NIL en caso contrario. POS es una posición del tablero dada por una lista con valor de coordenada X y valor de coordenada Y. DIR ha de ser un número entero (dirección) entre 0 y 3 (norte, este, sur y oeste respectivamente) o NIL si se quiere consultar la posición POS."
  (member (envio:consulta-tipo (envio:consulta-pos-tab pos dir)) '(envio:planta nil)))

(defconstant dir-inc
  (make-array '(4 2)
	      :initial-contents
	      '((1 -1)    ; Norte
		(0 +1)    ; Este
		(1 +1)    ; Sur
		(0 -1)))  ; Oeste
  "Array para transformar de dirección a número de coordenada a modificar y valor de incremento.")

(defmacro dir2dim-inc (dir)
  `(aref dir-inc ,dir 0))

(defmacro dir2val-inc (dir)
  `(aref dir-inc ,dir 1))

(defun pos-inc-dir (pos dir)
  "Calcula nueva lista con la posición resultante de desplazarse en la dirección DIR respecto a POS."
  (let ((pos-inc (copy-list pos)))
    (setf (nth (dir2dim-inc dir) pos-inc)
	  (+ (nth (dir2dim-inc dir) pos-inc)
	     (dir2val-inc dir)))
    pos-inc))

(defun distancia-city (pos1 pos2)
  "Calcula la distancia city block entre POS1 y POS2."
  (+ (abs (- (first pos1) (first pos2)))
     (abs (- (second pos1) (second pos2)))))


;;
;; Funciones para generar sistema de producción
;; --------------------------------------------

(defun estados-iniciales (id)
  "Genera lista de estados iniciales del jugador con identificador ID."
  (let ((nue-est (make-estado :camino (list (envio:consulta-pos-jug id)))))
    (if *f-heur*
	(setf (estado-f-heur nue-est) (funcall *f-heur* nue-est)))
    (list nue-est)))

(defun precond-regla-dir (est dir)
  "Generador de precondicion de regla de mover el estado EST en dirección DIR."
  (let ((pos-est (first (estado-camino est))))
    ; La nueva posición no pertenece al propio camino
    (and (not (member (pos-inc-dir pos-est dir)
		      (rest (estado-camino est))
		      :test #'equal))
	 ; Y la nueva posición se puede pisar
	 (pos-posible-p pos-est dir))))

(defun accion-regla-dir (est dir)
  "Generador de acción de regla de mover el estado EST en dirección DIR."
  (let ((nue-est (make-estado :camino (cons (pos-inc-dir (first (estado-camino est)) dir)
					    (estado-camino est))
			      :lon-camino (1+ (estado-lon-camino est))
			      :dir-ini (or (estado-dir-ini est) dir))))
    ; Asignar valor heurísticos, si se requiere
    (if *f-heur*
	(setf (estado-f-heur nue-est) (funcall *f-heur* nue-est)))
    nue-est))

(defun print-camino (camino)
  "Solicita imprimir la lista de posiciones dada en CAMNIO a los agentes capaces de acpetar solicitudes tipo :punteo."
  (envio:broadcast :punteo (reverse camino)))


;;
;; Inicializador de agentes de búsqueda de camino
;; ----------------------------------------------

(defun aluXXXX-camino-ini (id tipo-juego)
  "Inicialización genérica de agente que buscan camino."
  (declare (ignore id tipo-juego))          ; Ignorar parámetros no usados
  (if (pt-util:usa-sistema 'camino-planta)   ; Si no existían ya las reglas
      (progn
	; Insertar las nuevas reglas
	(pt-util:regla #'(lambda (est) (precond-regla-dir est 3))  ; Regla Oeste
		       #'(lambda (est) (accion-regla-dir est 3)))
	(pt-util:regla #'(lambda (est) (precond-regla-dir est 1))  ; Regla Este
		       #'(lambda (est) (accion-regla-dir est 1)))
	(pt-util:regla #'(lambda (est) (precond-regla-dir est 0))  ; Regla Norte
		       #'(lambda (est) (accion-regla-dir est 0)))
	(pt-util:regla #'(lambda (est) (precond-regla-dir est 2))  ; Regla Sur
		       #'(lambda (est) (accion-regla-dir est 2)))))
  t)    ; Devolver T siempre


;;
;; Agente búsqueda de camino en anchura
;; ------------------------------------

(defun estado-objetivo-p (est)
  "Chequea si un estado EST es estado objetivo o NIL."
  ; Si se ha acabado el tiempo o no se ha encontrado solución
  (or (pt-util:consulta-contador) (null est)
      ; O ya se ha alcalnzado una planta
      (eq 'envio:planta
	  (envio:consulta-tipo (envio:consulta-pos-tab (first (estado-camino est)))))))

(defun aluXXXX-camino-ba (id)
  "Agente jugador PlanTron que busca camino en anchura a planta."
  ; Iniciar contador de tiempo transcurrido
  (pt-util:inicia-contador)
  ; Usar sistema de producción de la presa
  (pt-util:usa-sistema 'camino-planta)
  ; Cargar heurística a utilizar (sin heurística)
  (setf *f-heur* nil)
  ; Lanzar búsqueda
  (let ((est-sol
	 (pt-util:generic-search
	  (estados-iniciales id)
	  #'estado-objetivo-p
	  #'pt-util:hijos
	  :display-fn #'null
	  :merge-fn   #'(lambda (new-states)
			  (nconc pt-util:*search-queue* new-states))
	  :dequeue-fn #'(lambda () (pop pt-util:*search-queue*)))))
    ; Preguntar por condiciones de parada
    (cond
     ((or (null est-sol)                    ; No encontrada solución (encerrado)
	  (null (estado-dir-ini est-sol)))  ; o estado inicial es solución
      (envio:mensaje "Lanzo cobarde")
      (pt-agentes-simples::agente-cobarde id))    ; Como agente-cobarde

     
     ((pt-util:consulta-contador) ; Agotado el tiempo de búsqueda
      (envio:mensaje "Agotado el tiempo de búsqueda")
      (pt-agentes-simples::agente-espiral id))    ; Como agente-espiral
     
     (t                           ; Encontrado camino
      (envio:mensaje "Encontrado camino BA")
      (print-camino (estado-camino est-sol))
      (estado-dir-ini est-sol)))))                ; Mover según estado solución


;;
;; Agente búsqueda de camino A*
;; ----------------------------

(defun f-dist-planta (est)
  "Función heurística subdividida con coste del camino acumulado + distancia city block a planta más cercana"
  (+ (estado-lon-camino est)
     (second (first (pt-util:planta-cerca (first (estado-camino est)))))))

(defun estado-objetivo-a-p (est)
  "Chequea si un estado EST es estado objetivo o NIL. Aprovecha el campo F-HEUR del estado."
  ; Si se ha acabado el tiempo o no se ha encontrado solución
  (or (pt-util:consulta-contador) (null est)
      ; O ya se ha alcalnzado una planta
      (>= (estado-lon-camino est) (estado-f-heur est))))

(defun igual-estado (est1 est2)
  "Comprueba si el estado EST1 es igual al EST2"
  ; Considerar dos estados iguales si su camino
  ; acaba en la misma posición
  (equal (first (estado-camino est1))
         (first (estado-camino est2))))

(defun aluXXXX-camino-a (id)
  "Agente jugador PlanTron que busca con algoritmo A camino a planta."
  ; Iniciar contador de tiempo
  (pt-util:inicia-contador)
  ; Usar sistema de producción de la presa
  (pt-util:usa-sistema 'camino-planta)
  ; Inicializar variables para heurísticas del A*
  (setf *f-heur* #'f-dist-planta)                     ; Fun. f' para reglas
  (setf pt-util:*coste-acum-fn* #'estado-lon-camino)  ; Fun. g  para búsqueda
  (setf pt-util:*eval-fn* #'estado-f-heur)            ; Fun. f' para búsqueda
  (setf pt-util:*igual-est-fn* #'igual-estado)        ; Pred. estados iguales
  ; Lanzar búsqueda
  (let ((est-sol
	 (pt-util:generic-search
	  (estados-iniciales id)
	  #'estado-objetivo-a-p
	  #'pt-util:hijos
	  :display-fn #'null
	  :merge-fn   #'pt-util:priority-a-merge-fn
	  :dequeue-fn #'pt-util:priority-a-dequeue-fn)))
    ; Preguntar por condiciones de parada
    (cond
     ((or (null est-sol)                    ; No encontrada solución (encerrado)
	  (null (estado-dir-ini est-sol)))  ; o estado inicial es solución
      (envio:mensaje "Lanzo cobarde")
      (pt-agentes-simples::agente-cobarde id))    ; Como agente-cobarde

     ((pt-util:consulta-contador) ; Agotado el tiempo de búsqueda
      (envio:mensaje "Agotado el tiempo de búsqueda")
      (pt-agentes-simples::agente-espiral id))    ; Como agente-espiral
     
     (t                           ; Encontrado camino
      (envio:mensaje "Encontrado camino A*")
      (print-camino (estado-camino est-sol))
      (estado-dir-ini est-sol)))))                ; Mover según estado solución


;;
;; Definiciones de agentes presentes en este package
;; -------------------------------------------------

(pt:defagente aluXXXX-camino-ba
	      :turno #'aluXXXX-camino-ba
	      :inicio #'aluXXXX-camino-ini
	      :use '("CONTADOR" "SHELL" "GENERIC-SEARCH")
	      :documentation "Agente jugador PlanTron que busca camino en anchura a planta.")

(pt:defagente aluXXXX-camino-a
	      :turno #'aluXXXX-camino-a
	      :inicio #'aluXXXX-camino-ini
	      :use '("CONTADOR" "SHELL" "GENERIC-SEARCH" "A_STAR" "PLANTA-CERCA")
	      :documentation "Agente jugador PlanTron que busca camino con algoritmo A a planta.")
